require 'inflecto'

module Sire
  module EntityBuilder
    class << self
      def build(hash)
        create_lookup unless lookup_up_to_date?
        parse_values(hash)
      end

      def nestables
        @nestables ||= []
      end

      def primitives
        @primitves ||= [String, Numeric, Rational, Complex, Integer, Fixnum, Float, Bignum,
                        TrueClass, FalseClass, Symbol, Time, Regexp, NilClass]
      end

      private

      def create_lookup
        @nestable_lookup = Hash[nestables.map { |n| [to_sym(n), n] }]
      end

      def lookup_up_to_date?
        @nestable_lookup.respond_to?(:values) && @nestable_lookup.values == nestables
      end

      def parse_values(hash)
        Hash[hash.map { |k, v| nested(k, v) || primitive(k, v) || invalid(k) }]
      end

      def nested(k, v)
        return unless nested_class = @nestable_lookup[k]
        [k, nested_class.new(v)]
      end

      def primitive(k, v)
        [k, v] if primitives.include?(v.class)
      end

      def invalid(k)
        fail InvalidValueError, "#{k} is invalid. Accepted types: #{primitives | nestables}"
      end

      InvalidValueError = Class.new(StandardError)

      def to_sym(klass)
        Inflecto.underscore(Inflecto.demodulize(klass.to_s)).to_sym
      end
    end
  end
end
