module Sire
  class Struct
    include Enumerable

    class << self
      attr_reader :attributes

      def define(*attributes)
        Class.new(self) do
          attr_accessor(*attributes)
          private(*attributes.map { |a| "#{a}=" })
          @attributes = [*attributes].freeze
        end
      end

      def [](hash = {})
        new(hash)
      end
    end

    def initialize(hash = {})
      hash.each { |k, v| send("#{k}=", v) }
      freeze
    end

    def merge(other)
      other_hash = other.respond_to?(:to_h) ? other.to_h : other
      self.class.new(to_h.merge(other_hash))
    end
    alias_method :update, :merge

    def attributes
      self.class.attributes
    end

    def each
      attributes.each { |a| yield a, send(a) }
    end

    def ==(other)
      self.class == other.class && to_h == other.to_h
    end

    def to_s
      "#{self.class}[#{map { |k, v| "#{k}: #{v.inspect}" }.join(', ') }]"
    end

    def inspect
      to_s
    end

    def to_h
      Hash[to_a]
    end
  end
end
